from unittest import TestCase
from src.smalltools_st.behavior.suitable_class_finder import is_concrete, concrete_subclasses, SuitableClassFinder
from .utils import *


class SuitableClassFinderTestCase(TestCase):

    def test_is_concrete(self):
        self.assertFalse(is_concrete(Grandfather))
        self.assertTrue(is_concrete(Aunt))
        self.assertFalse(is_concrete(Uncle))
        self.assertTrue(is_concrete(Cousin))
        self.assertFalse(is_concrete(Father))
        self.assertTrue(is_concrete(Brother))
        self.assertTrue(is_concrete(Sister))
        self.assertTrue(is_concrete(Me))

    def test_concrete_subclasses(self):
        self.assertSetEqual(set(concrete_subclasses(Grandfather)), set(Aunt, Cousin, Brother, Sister, Me))
        self.assertSetEqual(set(concrete_subclasses(Aunt)), set())
        self.assertSetEqual(set(concrete_subclasses(Uncle)), set(Cousin))
        self.assertSetEqual(set(concrete_subclasses(Cousin)), set())
        self.assertSetEqual(set(concrete_subclasses(Father)), set(Brother, Sister, Me))
        self.assertSetEqual(set(concrete_subclasses(Brother)), set())
        self.assertSetEqual(set(concrete_subclasses(Sister)), set())
        self.assertSetEqual(set(concrete_subclasses(Me)), set())

    def test_can_handle(self):
        # ONLY concrete subclasses can handle the suitable object
        self.assertEqual(Me, SuitableClassFinder(Relative).suitable_for("John"))
    
        # NO MORE THAN ONE concrete subclass can handle the suitable object
        with self.assertRaises(ValueError):
            SuitableClassFinder(Relative).suitable_for("Jenny")
                
        # JUST ONE concrete subclass should handle the suitable object
        with self.assertRaises(ValueError):
            SuitableClassFinder(Relative).suitable_for("Bob")

    def test_defaultSubclass(self):
        # You can define a default class
        self.assertEqual(Nobody, SuitableClassFinder(Relative).suitable_for("Bob", defaultSubclass=Nobody))

    def test_suitableMethod(self):
        # You can specify an explicit canHandle method (useful for multiple positional arguments)
        self.assertEqual(Sister, SuitableClassFinder(Relative).suitable_for("Jenny", 14, suitableMethod='canHandleNameAndAge'))



is_concrete = lambda subclass: len(subclass.__subclasses__()) == 0


def concrete_subclasses(a_class, accumulator=[]):
    """For this package, concrete class is any with no subclasses.
    
    Parameters:
    - :a_class: : Any class object
    - :accumulator: (optional): A list with all concrete subclasses partially found for :a_class:

    Returns:
    - :accumulator: : List of concrete subclasses
    """
    for subclass in a_class.__subclasses__():
        accumulator.append(subclass) if is_concrete(subclass) else concrete_subclasses(subclass, accumulator)
    return accumulator


class SuitableClassFinder():
    """Given the hierarchy of an abstract class, it detects the appropriate 
    concrete subclass (deterministically) that satisfies certain attributes 
    obtained as a parameter.
    Useful for implementing the Strategy design pattern.
    """


    def __init__(self, abstract_class):
        """Initialization of the subclass finder.

        Parameter:
        - :abstract_class: : Any class object
        """
        self.abstract_class = abstract_class
        super().__init__()


    def suitable_for(self, *suitable_object, default_subclass=None, suitable_method='can_handle'):
        """Finds the concrete subclass that satisfies the conditions modeled 
        with the :suitable_object: and the :suitable_method:

        Parameters:
        - :*suitable_object: : Positional arguments of any kind
        - :default_subclass: (optional): Its the object to be returned when no subclass is found. Keep
        it as :None: to ensure a match or an exception.
        - :suitable_method: (optional): Its the boolean method that tests each possible subclass against 
        the :suitable_object:. It must be implemented in the abstract class or in its all subclasses. Keep
        it as :can_handle: for standardization.

        Returns:
        If it success, a unique subclass of the abstract class provided, which satisfies the :suitable_method:
        using the :suitable_object: as parameters.
        """
        all_subclasses = concrete_subclasses(self.abstract_class)
        filtered_subclasses = \
            [subclass for subclass in all_subclasses 
                if (getattr(subclass, suitable_method)(*suitable_object))]

        subclasses_amount = len(filtered_subclasses)
        if (subclasses_amount == 0) and (default_subclass is not None):
            return default_subclass
        elif (subclasses_amount == 1):
            return filtered_subclasses[0]
        elif (subclasses_amount > 1):
            raise ValueError(f"Many subclasses can handle the suitable object '{suitable_object}'")
        else:
            raise ValueError(f"No subclass can handle the suitable object '{suitable_object}'")